// Copyright (C) 2016-2024  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

// Package zodbpickle provides pickling facility tailored to ZODB/go needs.
//
// Contrary to zodbpickle/py it merely wraps [github.com/kisielk/og-rek] without
// forking code from anywhere.
//
// Use [NewPickler]/[NewUnpickler] to create pickle encoder and decoder
// correspondingly. The decoder uses StrictUnicode=y mode to prevent loading
// *STRING and *UNICODE opcodes as the same type "string" on Go side and thus
// loosing information. The encoder uses StrictUnicode=y mode as well for
// symmetry so that decode/encode cycle is identity. As the result
//
//   - py bytestring (str from py2) is represented as Go type ogórek.ByteString;
//   - py unicode (unicode from py2 and str from py3) is represented as Go type string;
//   - py bytes (bytes from py3 and zodbpickle.binary from py2) is represented as Go type ogórek.Bytes.
//
// At application level utilities like ogórek.AsBytes and ogórek.AsString are
// handy to work with unpickled data for pickles generated by either py2 or py3.
//
// The decoder also uses PyDict=y mode to generally adhere to Python semantic
// and to allow programs to access unpickled dictionaries via string/bytes keys
// following py3 model, while still being able to handle dictionaries
// generated from under py2.
//
// The encoder emits pickles with protocol=3 to natively support all py bytes
// and strings types, and to stay interoperable with both py2 and py3: both ZODB4
// and ZODB5 use zodbpickle which supports protocol=3 on py2 and py3, and ZODB5
// actually saves pickles with protocol=3 on both py2 and py3 starting from
// ZODB 5.4 . Pickles saved with protocol=3 are thus universally readable.
//
// See package [github.com/kisielk/og-rek] for details of pickling/unpickling on Go side.
package zodbpickle

import (
	"encoding/binary"
	"fmt"
	"io"

	pickle "github.com/kisielk/og-rek"
)

// NewPickler creates new pickle encoder.
//
// The encoder adheres to semantic explained in top-level documentation of
// zodbpickle package.
//
// getref, if !nil, will be used by the encoder to see if a referenced object
// should be represented by persistent reference in generated pickle stream.
// See documentation of ogórek.EncoderConfig.PersistentRef for details.
func NewPickler(w io.Writer, getref func(obj any) *pickle.Ref) *pickle.Encoder {
	return pickle.NewEncoderWithConfig(w, &pickle.EncoderConfig{
		Protocol:      3,    // see top-level doc
		StrictUnicode: true, // see top-level doc
		PersistentRef: getref,
	})
}

// NewUnpickler creates new pickle decoder.
//
// The decoder adheres to semantic explained in top-level documentation of
// zodbpickle package.
//
// loadref, if !nil, will be used by the decoder to handle persistent
// references. See documentation of ogórek.DecoderConfig.PersistentLoad for details.
func NewUnpickler(r io.Reader, loadref func(ref pickle.Ref) (any, error)) *pickle.Decoder {
	return pickle.NewDecoderWithConfig(r, &pickle.DecoderConfig{
		StrictUnicode:  true, // see top-level doc
		PyDict:         true, // see top-level doc
		PersistentLoad: loadref,
	})
}


// ---- pack/unpack uint64 into [8]bytes ----

// Pack64 packs uint64 into big-endian bytes.
func Pack64(x uint64) pickle.Bytes {
	var b [8]byte
	binary.BigEndian.PutUint64(b[:], x)
	return pickle.Bytes(b[:])
}

// Unpack64 tries to extract [8](bytestr|bytes) from unpickled value as big-endian uint64.
func Unpack64(x interface{}) (uint64, error) {
	s, err := pickle.AsBytes(x)
	if err != nil {
		return 0, err
	}

	if len(s) != 8 {
		return 0, fmt.Errorf("expect [8]bytes; got [%d]bytes", len(s))
	}

	return binary.BigEndian.Uint64([]byte(s)), nil
}
