// Copyright (C) 2017-2024  Nexedi SA and Contributors.
//                          Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package fs1tools

//go:generate ./gen-testdata

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"strings"
	"testing"

	"lab.nexedi.com/kirr/neo/go/internal/xtesting"
	"lab.nexedi.com/kirr/neo/go/zodb/storage/fs1"

	"github.com/kylelemons/godebug/diff"
)

// ztestdataReg maintains registry of all entries under testdata/ .
var  ztestdataReg = xtesting.ZTestDataRegistry[struct{}]{}
type ZTestData    = xtesting.ZTestData[struct{}]


func loadFile(t *testing.T, path string) string {
	data, err := ioutil.ReadFile(path)
	if err != nil {
		t.Fatal(err)
	}
	return string(data)
}

func testDump(t *testing.T, dir fs1.IterDir, newd func() Dumper) {
	ztestdataReg.RunWithEach(t, func(t *testing.T, z *ZTestData) {
		_testDump(t, dir, newd, z)
	})
}

func _testDump(t *testing.T, dir fs1.IterDir, newd func() Dumper, z *ZTestData) {
	testv := []string{"1", "empty", "whiteout"}
	for _, tt := range testv {
		t.Run("db=" + tt, func(t *testing.T) {
			d   := newd()
			buf := bytes.Buffer{}

			err := Dump(&buf, "../"+z.Path(tt+".fs"), dir, d)
			if err != nil {
				t.Fatalf("%s: %v", d.DumperName(), err)
			}

			dumpOk := loadFile(t, z.Path(fmt.Sprintf("/%s.%s.ok", tt, d.DumperName())))

			// normalize dumpOk from py3 to py2 and go outputs:
			// bytestrings come as '...' on py2 and b'...' on py3
			if strings.HasPrefix(z.Kind, "py3") {
				sub := func(old, new string) {
					dumpOk = strings.ReplaceAll(dumpOk, old, new)
				}
				sub(`: b'`, `: '`)
				sub(`: b"`, `: "`)
				sub(`=b'`,  `='`)
				sub(`=b"`,  `="`)
			}

			if dumpOk != buf.String() {
				t.Errorf("%s: dump different:\n%v", d.DumperName(), diff.Diff(dumpOk, buf.String()))
			}
		})
	}
}

func newFsDump()  Dumper { return &DumperFsDump{} }
func newFsDumpv() Dumper { return &DumperFsDumpVerbose{} }
func newFsTail()  Dumper { return &DumperFsTail{Ntxn: 1000000} }

func TestFsDump(t *testing.T)	{ testDump(t, fs1.IterForward,  newFsDump) }
func TestFsDumpv(t *testing.T)	{ testDump(t, fs1.IterForward,  newFsDumpv) }
func TestFsTail(t *testing.T)	{ testDump(t, fs1.IterBackward, newFsTail) }

func BenchmarkTail(b *testing.B) {
	// FIXME small testdata/1.fs is not representative for benchmarking
	for i := 0; i < b.N; i++ {
		err := Dump(ioutil.Discard, "../testdata/1.fs", fs1.IterBackward, &DumperFsTail{Ntxn: 1000000})
		if err != nil {
			b.Fatal(err)
		}
	}
}
